/*
 * @COPYRIGHT@
 * 
 * x-kernel v3.3
 * 
 * Copyright (c) 1993,1991,1990,1996  Arizona Board of Regents
 * 
 * @COPYRIGHT@
 *
 * $RCSfile: scout_monitor.c,v $
 *
 * HISTORY
 * $Log: scout_monitor.c,v $
 * Revision 1.4  1996/01/30  22:49:51  slm
 * Updated copyright and version.
 *
 * Revision 1.3  1995/01/18  07:22:17  davidm
 * (scout_monitor): renamed p->abbrev to p->dim_abbrev.
 *
 * Revision 1.2  1994/11/21  18:15:57  davidm
 * Renamed GMON_PROF_* to SCOUT_MON_*.
 *
 * Revision 1.1  1994/11/20  06:40:04  davidm
 * Initial revision
 */
/*
 * Performance monitoring support.
 */
/*-
 * Copyright (c) 1983, 1992, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include <sys/types.h>
#include "scout_malloc.h"
#include "scout_monitor.h"
#include "scout_sync_arch.h"
#include "scout_time.h"
#define bfd_vma	unsigned long
#include "gmon.h"
#include "gmon_out.h"

/* structure emitted by "gcc -a": */
struct bb {
    long		zero_word;
    const char		*filename;
    long		*counts;
    long		ncounts;
    struct bb		*next;
    const unsigned long	*addresses;
};

extern void profil();
static struct bb *bb_head;

struct gmonparam _gmonparam = {
    SCOUT_MON_OFF, "seconds", 's', -1, 0,
};

/*
 * See profil(2) where this is described:
 */
static int	s_scale;
#define		SCALE_1_TO_1	0x10000L


static void
hist_startup(lowpc, highpc)
     u_long lowpc;
     u_long highpc;
{
    register int o;
    char *cp;
    struct gmonparam *p = &_gmonparam;

    /*
     * round lowpc and highpc to multiples of the density we're using
     * so the rest of the scaling (here and in gprof) stays in ints.
     */
    p->lowpc = ROUNDDOWN(lowpc, HISTFRACTION * sizeof(HISTCOUNTER));
    p->highpc = ROUNDUP(highpc, HISTFRACTION * sizeof(HISTCOUNTER));
    p->textsize = p->highpc - p->lowpc;
    p->kcountsize = p->textsize / HISTFRACTION;
    p->hashfraction = HASHFRACTION;
    p->fromssize = p->textsize / HASHFRACTION;
    p->tolimit = p->textsize * ARCDENSITY / 100;
    if (p->tolimit < MINARCS) {
	p->tolimit = MINARCS;
    } else if (p->tolimit > MAXARCS) {
	p->tolimit = MAXARCS;
    } /* if */
    p->tossize = p->tolimit * sizeof(struct tostruct);

    cp = scout_malloc(p->kcountsize + p->fromssize + p->tossize);
    if (!cp) {
	printf("monstartup: out of memory\n");
	return;
    } /* if */
    /*
     * Don't do this on a machine with virtual memory that returns
     * zeroed out bss pages---it will page in all memory.
     */
    bzero(cp, p->kcountsize + p->fromssize + p->tossize);
    p->tos = (struct tostruct *)cp;
    cp += p->tossize;
    p->kcount = (u_short *)cp;
    cp += p->kcountsize;
    p->froms = (u_short *)cp;

    p->tos[0].link = 0;

    o = p->highpc - p->lowpc;
    if (p->kcountsize < o) {
#ifdef SCOUT_FPU_SUPPORT
	s_scale = ((float)p->kcountsize / o ) * SCALE_1_TO_1;
#else
	/* avoid floating point operations */
	int quot = o / p->kcountsize;
		
	if (quot >= 0x10000)
	  s_scale = 1;
	else if (quot >= 0x100)
	  s_scale = 0x10000 / quot;
	else if (o >= 0x800000)
	  s_scale = 0x1000000 / (o / (p->kcountsize >> 8));
	else
	  s_scale = 0x1000000 / ((o << 8) / p->kcountsize);
#endif
    } else {
	s_scale = SCALE_1_TO_1;
    } /* if */
} /* hist_startup */


static void
write_hist(void *sl)
{
    const u_char tag = GMON_TAG_TIME_HIST;
    struct gmonparam *p = &_gmonparam;
    struct gmon_hist_hdr thdr;
    int size;

    if (p->kcountsize > 0) {
	size = p->kcountsize / sizeof(HISTCOUNTER);
	if (p->rate < 0) {
	    p->rate = scout_ticks_per_second;
	} /* if */
	bcopy(&p->lowpc, &thdr.low_pc, sizeof(thdr.low_pc));
	bcopy(&p->highpc, &thdr.high_pc, sizeof(thdr.high_pc));
	bcopy(&size, &thdr.hist_size, sizeof(thdr.hist_size));
	bcopy(&p->rate, &thdr.prof_rate, sizeof(thdr.prof_rate));
	strcpy(thdr.dimen, p->dimen);
	thdr.dimen_abbrev = p->dim_abbrev;

	slog_write(sl, &tag, sizeof(tag));
	slog_write(sl, &thdr, sizeof(thdr));
	slog_write(sl, p->kcount, p->kcountsize);
    } /* if */
} /* write_hist */


static void
write_call_graph(void *sl)
{
    const u_char tag = GMON_TAG_CG_ARC;
    struct gmon_cg_arc_record raw_arc;
    int from_index, to_index, from_len;
    u_long frompc;

    from_len = _gmonparam.fromssize / sizeof(*_gmonparam.froms);
    for (from_index = 0; from_index < from_len; ++from_index) {
	if (_gmonparam.froms[from_index] == 0) {
	    continue;
	} /* if */

	frompc = _gmonparam.lowpc;
	frompc += from_index * _gmonparam.hashfraction
	  * sizeof(*_gmonparam.froms);
	for (to_index = _gmonparam.froms[from_index];
	     to_index != 0;
	     to_index = _gmonparam.tos[to_index].link)
	{
	    bcopy(&frompc, &raw_arc.from_pc, sizeof(raw_arc.from_pc));
	    bcopy(&_gmonparam.tos[to_index].selfpc, &raw_arc.self_pc,
		  sizeof(raw_arc.self_pc));
	    bcopy(&_gmonparam.tos[to_index].count, &raw_arc.count,
		  sizeof(raw_arc.count));

	    slog_write(sl, &tag, sizeof(tag));
	    slog_write(sl, &raw_arc, sizeof(raw_arc));
	} /* for */
    } /* for */
} /* write_call_graph */


static void
write_bb_counts(void *sl)
{
    struct bb *grp;
    const u_char tag = GMON_TAG_BB_COUNT;
    int ncounts;
    int i;

    /*
     * Write each group of basic-block info (all basic-blocks
     * in a compilation unit form a single group).
     */
    for (grp = bb_head; grp; grp = grp->next) {
	ncounts = grp->ncounts;
	slog_write(sl, &tag, sizeof(tag));
	slog_write(sl, &ncounts, sizeof(ncounts));
	for (i = 0; i < ncounts; ++i) {
	    slog_write(sl, &grp->addresses[i], sizeof(grp->addresses[0]));
	    slog_write(sl, &grp->counts[i], sizeof(grp->counts[0]));
	} /* for */
    } /* for */
} /* write_bb_counts */


static void
cleanup(void)
{
    struct gmonparam *p = &_gmonparam;
    const int version = GMON_VERSION;
    struct gmon_hdr ghdr;
    void *sl;

    sl = slog_open("gmon.out", "wb");
    if (!sl) {
	printf("cleanup: open of gmon.out failed!\n");
	scout_free(p->tos); p->tos = 0;
	return;
    } /* if */

    /* write gmon.out header: */
    bcopy(GMON_MAGIC, &ghdr.cookie[0], 4);
    bcopy(&version, &ghdr.version, sizeof(version));
    slog_write(sl, &ghdr, sizeof(ghdr));

    /* write PC histogram: */
    write_hist(sl);

    /* write call-graph: */
    write_call_graph(sl);

    /* write basic-block execution counts: */
    write_bb_counts(sl);

    slog_close(sl);
} /* cleanup */


/*
 * This function must not modify a1-a5.  As a last resort, write an
 * assembly stub that saves these registers before doing any work.
 */
void
__bb_init_func (struct bb *blocks)
{
    if (blocks->zero_word == 0 PREDICT_FALSE) {
	ScoutAtomicSequence(blocks->zero_word = 1;
			    blocks->next = bb_head;
			    bb_head = blocks);
    } /* if */
} /* __bb_init_func */


/*
 * Control profiling.
 *	Mode is a bit set of members SCOUT_MON_HIST, SCOUT_MON_CNT,
 *	and SCOUT_MON_BB.
 */
void
scout_monitor(int mode, void *lowpc, void *highpc)
{
    struct gmonparam *p = &_gmonparam;

    if (mode) {
	/* start */
	if (mode & SCOUT_MON_HIST) {
#ifdef SCOUT_PERFMON
	    strcpy(p->dimen, "events");
	    p->dim_abbrev = 'E';
	    p->rate = 1;
#endif
	    hist_startup((u_long)lowpc, (u_long)highpc);
	    scout_profile((char *)p->kcount, p->kcountsize, p->lowpc, s_scale);
	} /* if */
	if (mode & SCOUT_MON_BB) {
	    Scout_IPL old_ipl;
	    struct bb *ptr;

	    /* reset counters (bzero isn't profiled anyways): */
	    old_ipl = scout_swap_ipl(SCOUT_MAX_IPL);
	    for (ptr = bb_head; ptr != 0; ptr = ptr->next) {
		bzero(ptr->counts, ptr->ncounts * sizeof(ptr->counts[0]));
	    } /* for */
	    scout_swap_ipl(old_ipl);
	} /* if */
	p->state = mode;
    } else {
	/* stop all profiling: */
	scout_profile((char *)0, 0, 0, 0);
	if (p->state) {
	    p->state = SCOUT_MON_OFF;
	    cleanup();
	} /* if */
    } /* if */
} /* scout_monitor */

			/*** end of scout_monitor.c ***/
